#Kitty... A simple Panda3D framework
from pandac.PandaModules import loadPrcFileData

#using direct tools ... remove before distribution
DTOOLS = False
if DTOOLS:
    loadPrcFileData("", "want-tk #t")
#end direct tools

#Fullsceen mode option
FULLSCREEN = False
if FULLSCREEN:
    loadPrcFileData("", "fullscreen #t")
else:
    loadPrcFileData("", "fullscreen #f")

import direct.directbase.DirectStart
from direct.showbase.DirectObject import DirectObject
from direct.interval.LerpInterval import *
from pandac.PandaModules import *
from direct.gui.DirectGui import *
from direct.directtools.DirectGeometry import LineNodePath
from direct.gui.OnscreenText import OnscreenText
from pandac.PandaModules import TextNode
from panda3d.core import Vec3,Vec4
from pandac.PandaModules import OdeWorld, OdeSimpleSpace, OdeJointGroup, OdePlaneGeom
from pandac.PandaModules import OdeSphereGeom, OdeBoxGeom, OdeTriMeshData, OdeTriMeshGeom
from pandac.PandaModules import Quat, OdeBody, OdeMass

import sys, math, random




#***********************************************************
class World(DirectObject):
    #-------------------------------------------------------
    def __init__(self):
        self.models = {}      #dictionary of models
        self.bodies = {}      #dictionary of ode bodies
        self.textures = {}    #dic
        self.keys = {}
        self.changeQuat = Quat()
        self.setWindow("Kitty00", (0,0,0))
        self.SetupOdeWorld()
        self.setLight()
        self.setCameraLookat((0,-200,200),(0,0,0)) #default
        self.setupDefaults()
        self.loadModels()
        self.LoadTerrain("terrain","models/100xtestterrain","models/512brownterrain.jpg",(0,0,0))
        self.SetBodies()
        self.setupKeys()
        self.setupAnimations()


    #-------------------------------------------------------
    def SetupOdeWorld(self):
        self.odeworld = OdeWorld()
        self.odeworld.setGravity(0, 0, -9.81)
        self.odeworld.initSurfaceTable(1)
        self.odeworld.setSurfaceEntry(0, 0, 150, 0.0, 9.1, 0.9, 0.00001, 0.0, 0.002)
        self.deltaTimeAccumulator = 0.0
        self.stepSize = 1.0 / 90.0
        #self.space = OdeSimpleSpace()
        self.space = OdeHashSpace()
        self.space.setAutoCollideWorld(self.odeworld)
        self.contactgroup = OdeJointGroup()
        self.space.setAutoCollideJointGroup(self.contactgroup)


    #-------------------------------------------------------
    def SetBody(self, name, size, mass, type):
        body = OdeBody(self.odeworld)
        M = OdeMass()
        M = OdeMass()
        if type == "sphere":
            M.setSphere(mass, size)
        body.setPosition(self.models[name].getPos(render))
        body.setQuaternion(self.models[name].getQuat(render))

        if type == "sphere":
            geom = OdeSphereGeom(self.space, size)
        geom.setCollideBits(BitMask32(0x00000002))
        geom.setCategoryBits(BitMask32(0x00000001))
        geom.setBody(body)

        self.bodies[name] = body

    #-------------------------------------------------------
    def LoadTerrain(self,name,model,texture,position):
        self.loadModelAndTexture(name,model,texture,position,1)
        self.envCollideTrimesh = OdeTriMeshData(self.models[name], True)
        self.envCollideGeom = OdeTriMeshGeom(self.space, self.envCollideTrimesh)
        self.envCollideGeom.setPosition(self.models[name].getPos())
        self.envCollideGeom.setCategoryBits(BitMask32.allOn())
        self.envCollideGeom.setCollideBits(BitMask32.allOn())

    #-------------------------------------------------------
    def debug(self,text):
        txt = str(text)
        self.debugtext.setText(txt)

    #-------------------------------------------------------
    def setupDefaults(self):
        self.BLACK = Vec4(0,0,0,1)
        self.WHITE = Vec4(1,1,1,1)
        self.font = loader.loadFont("models/cmss12")
        self.debugtext = OnscreenText(text="Debug: ",
                              style=1, fg=(1,1,1,1), font = self.font,
                              pos=(-0.60, 0.95), scale = .07)
        self.setupProgramVariables()

    #-------------------------------------------------------
    def setupProgramVariables(self):
        self.distance = 0
        self.cameraPos = (0,-100,50)
        self.mouseX = 0
        self.mouseY = 0
        self.mouseOX = 0
        self.mouseOY = 0
        self.mouseKey = 0   #no button pressed - default
        self.mouseWheel = 0 #no wheel turn - default
        self.zoomScale = 30.0
        self.rotationXDelta = 1
        self.rotationYDelta = 0.5

    #-------------------------------------------------------
    def setWindow(self, title, bgcolor):
        base.disableMouse() #detatch camera from mouse control
        base.setBackgroundColor(*bgcolor)
        winProps = WindowProperties()
        winProps.setTitle(title)
        base.win.requestProperties(winProps)

    #-------------------------------------------------------
    def setLight(self):
        plight = PointLight('plight')
        plight.setColor(VBase4(1, 1, 1, 1))
        plnp = render.attachNewNode(plight)
        plnp.setPos(10, 0, 20)
        render.setLight(plnp)
        alight = AmbientLight('alight')
        alight.setColor(VBase4(0.3, 0.3, 0.3, 1))
        alnp = render.attachNewNode(alight)
        render.setLight(alnp)

    #-------------------------------------------------------
    def setupKeys(self):
        self.accept("escape", sys.exit)
        self.accept('mouse1',self.MouseLeftKey)
        self.accept('mouse2',self.MouseMiddleKey)
        self.accept('mouse3',self.MouseRightKey)
        self.accept('mouse1-up',self.MouseLeftKeyUp)
        self.accept('mouse2-up',self.MouseMiddleKeyUp)
        self.accept('mouse3-up',self.MouseRightKeyUp)
        self.accept('wheel_up',self.MouseWheelUp)
        self.accept('wheel_down',self.MouseWheelDown)

    #-------------------------------------------------------
    def MouseWheelUp(self):
        self.mouseWheel = 1

    #-------------------------------------------------------
    def MouseWheelDown(self):
        self.mouseWheel = 2

    #-------------------------------------------------------
    def MouseLeftKey(self):
        self.mouseKey = 1

    #-------------------------------------------------------
    def MouseMiddleKey(self):
        self.mouseKey = 2

    #-------------------------------------------------------
    def MouseRightKey(self):
        self.mouseKey = 3

    #-------------------------------------------------------
    def MouseLeftKeyUp(self):
        self.mouseKey = 0

    #-------------------------------------------------------
    def MouseMiddleKeyUp(self):
        self.mouseKey = 0

    #-------------------------------------------------------
    def MouseRightKeyUp(self):
        self.mouseKey = 0

    #-------------------------------------------------------
    def setKey(self, key, val):
        self.keys[key] = val

    #-------------------------------------------------------
    def setCamera(self, position, rotation):
        self.setCameraPosition(position)
        self.setCameraRotation(rotation)

    #-------------------------------------------------------
    def setCameraLookat(self, position, target):
        self.setCameraPosition(position)
        self.setCameraTarget(target)

    #-------------------------------------------------------
    def setCameraPosition(self, position):
        camera.setPos (*position)
        self.cameraPos = position

    #-------------------------------------------------------
    def GetCameraPosition(self):
        self.cameraPos = camera.getPos()

    #-------------------------------------------------------
    def setCameraRotation(self, rotation):
        camera.setHpr (*rotation)

    #-------------------------------------------------------
    def setCameraTarget(self, target):
        base.camera.lookAt(*target)

    #-------------------------------------------------------
    def setCameraToModel(self,model):
        camera.reparentTo(self.models[model])

    #-------------------------------------------------------
    def loadTexture(self, name, texturefile):
        texture = loader.loadTexture(texturefile)
        self.textures[name] = texture

    #-------------------------------------------------------
    def setTexture(self, modelname, texturename):
        self.models[modelname].setTexture(self.textures[texturename],1)

    #-------------------------------------------------------
    def loadModelAndTexture(self,name,modelfile,texturefile,position,scale):
        self.loadModel(name,modelfile,position,scale)
        self.loadTexture(name,texturefile)
        self.setTexture(name,name)

    #-------------------------------------------------------
    def loadModel(self, name, modelfile, position, scale):
        model = loader.loadModel(modelfile)
        model.reparentTo(render)
        model.setScale(scale)
        model.setPos(*position)
        model.setTag("name",name) #using tag to store Meta info
        self.models[name] = model

    #-------------------------------------------------------
    def load2dModel(self, name, modelfile, position, scale):
        model = loader.loadModel(modelfile)
        model.reparentTo(render2d)
        model.setScale(scale)
        model.setPos(*position)
        model.setTag("name",name) #using tag to store Meta info
        self.models[name] = model

    #-------------------------------------------------------
    def load2dModelAndTexture(self,name,modelfile,texturefile,position,scale):
        self.load2dModel(name,modelfile,position,scale)
        self.loadTexture(name,texturefile)
        self.setTexture(name,name)

    #-------------------------------------------------------
    def loadModels(self):
        self.loadModel("pointer","models/ball-1m",(0,0,0),0.0001) #1mm dummy pointer
        self.loadModelAndTexture("sky","models/solar_sky_sphere","models/stars_1k_tex.jpg",(0,0,0),200)
        self.loadModelAndTexture("ball","models/ball-1m","models/red_txt.jpg",(0,0,30),3)

    #-------------------------------------------------------
    def SetBodies(self):
        self.SetBody("ball",3,1,"sphere")


    #-------------------------------------------------------
    def setTag(self,model,tag,value):
        self.models[model].setTag(tag,str(value))

    #-------------------------------------------------------
    def getTag(self,model,tag):
        return self.models[model].getTag(tag)

    #-------------------------------------------------------
    def getTagFloat(self,model,tag):
        return float(self.models[model].getTag(tag))

    #-------------------------------------------------------
    def setupAnimations(self):
        taskMgr.add(self.task_doAnimations,"do aimations")
        taskMgr.add(self.task_MouseControl,"Mouse Camera Control")
        taskMgr.doMethodLater(1.0, self.task_Ode, "Ode Physics Loop")

    #-------------------------------------------------------
    def task_doAnimations(self,task):
        return task.cont

    #-------------------------------------------------------
    def task_Ode(self, task):
        self.space.autoCollide()
        self.deltaTimeAccumulator += globalClock.getDt()
        while self.deltaTimeAccumulator > self.stepSize:
            self.deltaTimeAccumulator -= self.stepSize
            self.odeworld.quickStep(self.stepSize)
            for o in self.bodies:
                self.models[o].setPosQuat(render, self.bodies[o].getPosition(),
                         Quat(self.bodies[o].getQuaternion()))
        self.contactgroup.empty()
        return task.cont

    #-------------------------------------------------------
    def task_MouseControl(self, task):
        if base.mouseWatcherNode.hasMouse():
            self.mouseX = base.mouseWatcherNode.getMouseX()
            self.mouseY = base.mouseWatcherNode.getMouseY()
        #self.debug(str(self.mouseX)+"  "+str(self.mouseOX))
        if self.mouseX < self.mouseOX:   #mouse moved left
            if self.mouseKey == 3:       #and right mouse button is in
                self.OrbitCameraLeft()
        if self.mouseX > self.mouseOX:   #mouse moved right
            if self.mouseKey == 3:       #and right mouse button is in
                self.OrbitCameraRight()
        if self.mouseY > self.mouseOY:   #mouse moved up
            if self.mouseKey == 3:       #and right mouse button is in
                self.OrbitCameraUp()
        if self.mouseY < self.mouseOY:   #mouse moved down
            if self.mouseKey == 3:       #and right mouse button is in
                self.OrbitCameraDown()
        if self.mouseWheel == 1:
            self.ZoomCameraIn()
        if self.mouseWheel == 2:
            self.ZoomCameraOut()
        self.mouseOX = self.mouseX       #save last mouse X position
        self.mouseOY = self.mouseY       #save last mouse Y position
        self.mouseWheel = 0             #reset mouse wheel
        self.setCameraLookat(self.cameraPos,(0,0,0))
        return task.cont

    #-------------------------------------------------------
    def OrbitCameraUp(self):
        self.GetCameraPosition()
        x,y,z = self.cameraPos
        v1 = Vec3(x,y,z)                        #vector for camera position
        v2 = Vec3(0,0,0)                        #vector for world center
        distance = self.GetDistance(v1,v2)      #get distance to camera
        self.models["pointer"].lookAt((x,y,z)) #set pointer to look at the camera
        rx,ry,rz = self.models["pointer"].getHpr() #get pointer orientation
        ry += self.rotationYDelta                   #rotate the pointer up
        self.models["pointer"].setHpr((rx,ry,rz))
        #use the pointer-forward to plit the new camera position
        forward = self.models["pointer"].getQuat().getForward()*distance
        self.setCameraLookat(forward,(0,0,0))

    #-------------------------------------------------------
    def OrbitCameraDown(self):
        self.GetCameraPosition()
        x,y,z = self.cameraPos
        v1 = Vec3(x,y,z)                        #vector for camera position
        v2 = Vec3(0,0,0)                        #vector for world center
        distance = self.GetDistance(v1,v2)      #get distance to camera
        self.models["pointer"].lookAt((x,y,z)) #set pointer to look at the camera
        rx,ry,rz = self.models["pointer"].getHpr() #get pointer orientation
        ry -= self.rotationYDelta                   #rotate the pointer down
        self.models["pointer"].setHpr((rx,ry,rz))
        #use the pointer-forward to plit the new camera position
        forward = self.models["pointer"].getQuat().getForward()*distance
        self.setCameraLookat(forward,(0,0,0))

    #-------------------------------------------------------
    def OrbitCameraLeft(self):
        self.GetCameraPosition()
        x,y,z = self.cameraPos
        v1 = Vec3(x,y,z)                        #vector for camera position
        v2 = Vec3(0,0,0)                        #vector for world center
        distance = self.GetDistance(v1,v2)      #get distance to camera
        self.models["pointer"].lookAt((x,y,z)) #set pointer to look at the camera
        rx,ry,rz = self.models["pointer"].getHpr() #get pointer orientation
        rx -= self.rotationXDelta                   #rotate the pointer left
        self.models["pointer"].setHpr((rx,ry,rz))
        #use the pointer-forward to plot the new camera position
        forward = self.models["pointer"].getQuat().getForward()*distance
        self.setCameraLookat(forward,(0,0,0))

    #-------------------------------------------------------
    def OrbitCameraRight(self):
        self.GetCameraPosition()
        x,y,z = self.cameraPos
        v1 = Vec3(x,y,z)                        #vector for camera position
        v2 = Vec3(0,0,0)                        #vector for world center
        distance = self.GetDistance(v1,v2)      #get distance to camera
        self.models["pointer"].lookAt((x,y,z)) #set pointer to look at the camera
        rx,ry,rz = self.models["pointer"].getHpr() #get pointer orientation
        rx += self.rotationXDelta                   #rotate the pointer right
        self.models["pointer"].setHpr((rx,ry,rz))
        #use the pointer-forward to plot the new camera position
        forward = self.models["pointer"].getQuat().getForward()*distance
        self.setCameraLookat(forward,(0,0,0))

    #-------------------------------------------------------
    def ZoomCameraIn(self):
        self.GetCameraPosition()
        x,y,z = self.cameraPos
        v1 = Vec3(x,y,z)                        #vector for camera position
        v2 = Vec3(0,0,0)                        #vector for world center
        distance = self.GetDistance(v1,v2)      #get distance to camera
        newdistance = distance/self.zoomScale*(self.zoomScale-1)
        self.models["pointer"].lookAt((x,y,z)) #set pointer to look at the camera
        #get the pointer-forward, and use it to get the new camera position
        forward = self.models["pointer"].getQuat().getForward()*newdistance
        self.setCameraPosition(forward)

    #-------------------------------------------------------
    def ZoomCameraOut(self):
        self.GetCameraPosition()
        x,y,z = self.cameraPos
        v1 = Vec3(x,y,z)                        #vector for camera position
        v2 = Vec3(0,0,0)                        #vector for world center
        distance = self.GetDistance(v1,v2)      #get distance to camera
        newdistance = distance/self.zoomScale*(self.zoomScale+1)
        self.models["pointer"].lookAt((x,y,z)) #set pointer to look at the camera
        #get the pointer-forward, and use it to get the new camera position
        forward = self.models["pointer"].getQuat().getForward()*newdistance
        self.setCameraPosition(forward)

    #-------------------------------------------------------
    def GetDistance(self, v1, v2):
        vv1 = Vec3(v1[0],v1[2],v1[2])
        vv2 = Vec3(v2[0],v2[2],v2[2])
        vv = v2-v1
        return vv.length()



#***********************************************************
if __name__ == '__main__':
    w = World()
    run()